"""Air uses custom response classes to improve the developer experience."""

from collections.abc import Mapping
from typing import override

from starlette.background import BackgroundTask
from starlette.datastructures import URL
from starlette.responses import (
    FileResponse as FileResponse,
    HTMLResponse as HTMLResponse,
    JSONResponse as JSONResponse,
    PlainTextResponse as PlainTextResponse,
    RedirectResponse as StarletteRedirectResponse,
    Response as Response,
    StreamingResponse as StreamingResponse,
)
from starlette.types import Send

from .tags import BaseTag


class AirResponse(HTMLResponse):
    """Response class to handle air.tags.Tags or HTML (from Jinja2)."""

    @override
    def render(self, tag: BaseTag | str) -> bytes:
        """Render Tag elements to bytes of HTML."""
        return super().render(str(tag))


TagResponse = AirResponse  # Alias for clarity


class SSEResponse(StreamingResponse):
    """Response class for Server Sent Events

    Example:

        # For tags
        import random
        from asyncio import sleep

        import air

        app = air.Air()


        @app.page
        def index():
            return air.layouts.mvpcss(
                air.Script(src="https://unpkg.com/htmx-ext-sse@2.2.1/sse.js"),
                air.Title("Server Sent Event Demo"),
                air.H1("Server Sent Event Demo"),
                air.P("Lottery number generator"),
                air.Section(
                    hx_ext="sse",
                    sse_connect="/lottery-numbers",
                    hx_swap="beforeend show:bottom",
                    sse_swap="message",
                ),
            )

        async def lottery_generator():
            while True:
                lottery_numbers = ", ".join([str(random.randint(1, 40)) for x in range(6)])
                # Tags work seamlessly
                yield air.Aside(lottery_numbers)
                # As do strings. Non-strings are cast to strings via the str built-in
                yield "Hello, world"
                await sleep(1)


        @app.get("/lottery-numbers")
        async def get():
            return air.SSEResponse(lottery_generator())
    """

    media_type = "text/event-stream"

    async def stream_response(self, send: Send) -> None:
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            },
        )
        async for chunk in self.body_iterator:
            if not isinstance(chunk, bytes | memoryview):
                lines = list(str(chunk).splitlines())
                formatted = [f"data: {t}" for t in lines]
                data = "\n".join(formatted)
                chunk = f"event: message\n{data}\n\n"
                chunk = chunk.encode(self.charset)
            await send({"type": "http.response.body", "body": chunk, "more_body": True})

        await send({"type": "http.response.body", "body": b"", "more_body": False})


class RedirectResponse(StarletteRedirectResponse):
    def __init__(
        self,
        url: str | URL,
        status_code: int = 303,
        headers: Mapping[str, str] | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(url=url, status_code=status_code, headers=headers, background=background)



"""Air loves Jinja!

A common pattern is to use a Jinja template as the project base and then use Air Tags for individual content.
"""

import importlib
import inspect
from collections.abc import Callable, Sequence
from os import PathLike
from types import ModuleType
from typing import Any

import jinja2
from fastapi.templating import Jinja2Templates
from starlette.templating import _TemplateResponse

from .exceptions import RenderException
from .requests import Request


class JinjaRenderer:
    """Template renderer to make Jinja easier in Air.

    Args:
        directory: The template directory where Jinja templates for the project are stored.
        context_processors: A list of Jinja-style context processors, functions that automatically injects variables or functions into the template context so they're available in every rendered template without passing them explicitly.
        env: The env is the central Jinja object that holds configuration, filters, globals, and template loading settings, and is responsible for compiling and rendering templates.

    Example:

        # Instantiate the render callable
        jinja = JinjaRenderer('templates')

        # Use for returning Jinja from views
        @app.get('/')
        async def home(request: Request):
            return jinja(
                request,
                'home.html',
                context={'id': 5}
             )

            # Can also pass in kwargs, which will be added to the context:
            return jinja(
                request,
                'home.html',
                name='Parmesan'
            )

            # Will render Air Tags sent into Jinja context
            return jinja(
                request,
                'home.html',
                content=air.Article(air.P('Cheddar'))
            )

    """

    def __init__(
        self,
        directory: str | PathLike[str] | Sequence[str | PathLike[str]],
        context_processors: list[Callable[[Request], dict[str, Any]]] | None = None,
        env: jinja2.Environment | None = None,
    ) -> None:
        """Initialize with template directory path"""
        if context_processors is None:
            context_processors = []
        self.templates = Jinja2Templates(directory=directory, context_processors=context_processors, env=env)

    def __call__(
        self,
        request: Request,
        name: str,
        context: dict[Any, Any] | None = None,
        **kwargs: Any,
    ) -> _TemplateResponse:
        """Render template with request and context. If an Air Tag
        is found in the context, try to render it.
        """
        if context is None:
            context = {}
        if kwargs:
            context |= kwargs

        # Attempt to render any Tags in the contect
        context = {k: str(v) for k, v in context.items()}
        return self.templates.TemplateResponse(request=request, name=name, context=context)


class Renderer:
    """Template/Tag renderer to make composing pluggable functions easier.

    Args:
        directory: The template directory where Jinja templates for the project are stored.
        context_processors: A list of Jinja-style context processors, functions that automatically injects variables or functions into the template context so they're available in every rendered template without passing them explicitly.
        env: The env is the central Jinja object that holds configuration, filters, globals, and template loading settings, and is responsible for compiling and rendering templates.

    Example:

        import air

        app = air.Air()

        # Instantiate the render callable
        render = air.Renderer('templates')

        # Use for returning Jinja from views
        @app.get('/')
        async def home(request: Request):
            return render(
                name='home.html',
                request=request,
                context={'id': 5}
             )


            # Will render name of Air Tags
            return render(
                request,
                'components.home',
                context={'id': 5}
            )


            # Will render callables to HTML
            return render(
                air.layouts.mvpcss,
                air.Title("Test Page"),
                air.H1("Hello, World")
            )
    """

    def __init__(
        self,
        directory: str | PathLike[str] | Sequence[str | PathLike[str]],
        context_processors: list[Callable[[Request], dict[str, Any]]] | None = None,
        env: jinja2.Environment | None = None,
        package: str | None = None,
    ) -> None:
        """Initialize with template directory path"""
        if context_processors is None:
            context_processors = []
        self.templates = Jinja2Templates(directory=directory, context_processors=context_processors, env=env)
        self.package = package

    def __call__(
        self,
        name: str | Callable,
        *children: Any,
        request: Request | None = None,
        context: dict[Any, Any] | None = None,
        **kwargs: Any,
    ) -> str:
        """Render template with request and context. If an Air Tag
        is found in the context, try to render it.
        """
        context = self._prepare_context(context, kwargs)

        if callable(name):
            assert not isinstance(name, str)
            result = name(**context)
            if isinstance(result, str):
                return result
            if hasattr(result, "render"):
                return result.render()
            msg = "Callable in name arg must a string or object with a render method."
            raise TypeError(msg)

        assert isinstance(name, str)

        if name.endswith((".html", ".jinja")):
            return self._render_template(name, request, context)

        if "." in name:
            return self._render_tag_callable(name, children, request, context)

        msg = "No callable or Jinja template found."
        raise RenderException(msg)

    def _prepare_context(self, context: dict[Any, Any] | None, kwargs: dict[Any, Any]) -> dict[Any, Any]:
        """Prepare and merge context dictionaries."""
        if context is None:
            context = {}
        if kwargs:
            context |= kwargs
        return context

    def _render_template(self, name: str, request: Request | None, context: dict[Any, Any]) -> _TemplateResponse:
        """Render Jinja template with Air Tag support."""
        context = {k: str(v) for k, v in context.items()}
        return self.templates.TemplateResponse(request=request, name=name, context=context)

    def _render_tag_callable(self, name: str, args: tuple, request: Request | None, context: dict[Any, Any]) -> str:
        """Import and render a tag callable from module."""
        module_name, func_name = name.rsplit(".", 1)
        module = self._import_module(module_name)
        tag_callable = getattr(module, func_name)

        filtered_context = self._filter_context_for_callable(tag_callable, context, request)

        if filtered_context and args:
            return tag_callable(**filtered_context)
        return tag_callable(*args, **filtered_context)

    def _import_module(self, module_name: str) -> ModuleType:
        """Import module handling relative imports."""
        if module_name.startswith("."):
            return importlib.import_module(module_name, package=self.package)

        try:
            return importlib.import_module(module_name)
        except ModuleNotFoundError:
            return importlib.import_module(f".{module_name}", package=self.package)

    def _filter_context_for_callable(
        self, tag_callable: Callable, context: dict[Any, Any], request: Request | None
    ) -> dict[str, Any]:
        """Filter context to only include parameters expected by the callable."""
        sig = inspect.signature(tag_callable)
        filtered_context = {}

        for param_name in sig.parameters:
            if param_name in context:
                filtered_context[param_name] = context[param_name]

        if "request" in sig.parameters and request:
            filtered_context["request"] = request

        return filtered_context



"""Utilities for the Air Tag system."""

import html
from typing import Any, Final

EXTRA_FEATURE_PRETTY_ERROR_MESSAGE: Final = (
    "Extra feature 'pretty' is not installed. Install with: `uv add air[pretty]`"
)
FORMAT_HTML_ENCODING: Final = "unicode"
HTML_DOCTYPE: Final = "<!doctype html>"
DEFAULT_THEME: Final = "dracula"
PANEL_TITLE: Final = "Air → HTML"
PANEL_TITLE_STYLE: Final = "italic bold"
PANEL_BORDER_STYLE: Final = "bright_magenta"
SYNTAX_LEXER: Final = "html"


def clean_html_attr_key(key: str) -> str:
    """Clean up HTML attribute keys to match the standard W3C HTML spec.

    Args:
        key: An uncleaned HTML attribute key

    Returns:

        Cleaned HTML attribute key
    """
    # If a "_"-suffixed proxy for "class", "for", or "id" is used,
    # convert it to its normal HTML equivalent.
    key = {"class_": "class", "for_": "for", "id_": "id", "as_": "as"}.get(key, key)
    # Remove leading underscores and replace underscores with dashes
    return key.lstrip("_").replace("_", "-")


def pretty_format_html(
    source: str,
    *,
    with_body: bool = False,
    with_head: bool = False,
    with_doctype: bool = False,
) -> str:
    """
    Pretty-print HTML with pretty indentation and then unescape &lt;, &gt;, &amp; &#x27; in the final string.

    NOTE: This will also unescape inside attribute values, which can make the
    HTML invalid and unsafe. Use only if you fully trust the input/output.
    """

    return html.unescape(
        format_html(source, with_body=with_body, with_head=with_head, with_doctype=with_doctype, pretty=True)
    )


def format_html(
    source: str,
    *,
    with_body: bool = False,
    with_head: bool = False,
    with_doctype: bool = False,
    pretty: bool = False,
) -> str:
    try:
        from lxml import (
            etree,  # ty: ignore[unresolved-import]
            html as l_html,
        )
    except ModuleNotFoundError as exc:
        raise ModuleNotFoundError(EXTRA_FEATURE_PRETTY_ERROR_MESSAGE) from exc
    else:
        if with_body:
            source = l_html.document_fromstring(source, ensure_head_body=with_head)
        else:
            source = l_html.fromstring(source)
        if pretty:
            etree.indent(source)  # pretty indentation
        doctype = HTML_DOCTYPE if with_doctype else None
        return l_html.tostring(source, encoding=FORMAT_HTML_ENCODING, pretty_print=pretty, doctype=doctype)


def pretty_print_html(
    source: str,
    *,
    theme: str = DEFAULT_THEME,
) -> None:
    """
    Render HTML with syntax highlighting inside a compact, styled panel.

    Args:
        source: The HTML source to render.
        theme: Pygments style name; falls back (meaning: uses a safe default) if unknown.

    Raises:
        ModuleNotFoundError: If Rich (and its dependencies) are not available.
    """
    try:
        from rich import box
        from rich.console import Console
        from rich.panel import Panel
        from rich.syntax import Syntax
        from rich.text import Text
    except ModuleNotFoundError as exc:
        raise ModuleNotFoundError(EXTRA_FEATURE_PRETTY_ERROR_MESSAGE) from exc
    else:
        syntax = Syntax(source, SYNTAX_LEXER, theme=theme, line_numbers=True, indent_guides=True, word_wrap=True)
        title = Text(PANEL_TITLE, style=PANEL_TITLE_STYLE)
        panel = Panel.fit(
            syntax,
            box=box.HEAVY,
            border_style=PANEL_BORDER_STYLE,
            title=title,
        )
        console = Console()
        console.print(panel, soft_wrap=False)


def locals_cleanup(
    data: dict[str, Any],
    _skip: frozenset[str] = frozenset({"self", "children", "text_child", "args", "kwargs"}),
) -> dict[str, Any]:
    """Extract non-None attributes from locals() to merge with kwargs"""
    return {key: value for key, value in data.items() if value is not None and key[0] != "_" and key not in _skip}


class SafeStr(str):
    """A string subclass that doesn't trigger html.escape() when called by Tag.render()

    Example:
        sample = SafeStr('Hello, world')
    """



"""Air is proud to provide first class SVG support. The entire SVG specification
is supported.
"""

from typing import Any

from ..utils import locals_cleanup
from .base import AttributesType, BaseTag


class CaseTag(BaseTag):
    """This is for case-sensitive tags like those used in SVG generation."""

    @property
    def name(self) -> str:
        return self._name[0].lower() + self._name[1:]


class A(CaseTag):
    """Defines an SVG hyperlink

    Args:
        children: Tags, strings, or other rendered content.
        href: Hyperlink target URL.
        target: Where to display linked URL (_self|_parent|_top|_blank).
        download: Instructs browser to download instead of navigate.
        hreflang: Human language of the linked URL.
        ping: Space-separated list of URLs for tracking.
        referrerpolicy: Referrer policy when fetching the URL.
        rel: Relationship to target object.
        type: MIME type of linked URL.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        href: str | None = None,
        target: str | None = None,
        download: str | None = None,
        hreflang: str | None = None,
        ping: str | None = None,
        referrerpolicy: str | None = None,
        rel: str | None = None,
        type: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Animate(CaseTag):
    """Defines animation on an SVG element

    Args:
        children: Tags, strings, or other rendered content.
        attributeName: Target attribute to animate.
        attributeType: Type of target attribute.
        values: Values to animate through.
        dur: Total animation duration.
        repeatCount: Number of repetitions.
        repeatDur: Total duration for repeating.
        from_: Starting value (from is reserved).
        to: Ending value.
        by: Relative animation value.
        begin: Animation start time.
        end: Animation end time.
        calcMode: Interpolation mode (discrete|linear|paced|spline).
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        attributeName: str | None = None,
        attributeType: str | None = None,
        values: str | None = None,
        dur: str | None = None,
        repeatCount: str | float | None = None,
        repeatDur: str | None = None,
        from_: str | None = None,
        to: str | None = None,
        by: str | None = None,
        begin: str | None = None,
        end: str | None = None,
        calcMode: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class AnimateMotion(CaseTag):
    """Defines how an element moves along a motion path

    Args:
        children: Tags, strings, or other rendered content.
        path: Motion path using path syntax.
        keyPoints: Progress points along path (0-1 range).
        rotate: Rotation along path (Number|auto|auto-reverse).
        dur: Total animation duration.
        repeatCount: Number of repetitions.
        begin: Animation start time.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        path: str | None = None,
        keyPoints: str | None = None,
        rotate: str | float | None = None,
        dur: str | None = None,
        repeatCount: str | float | None = None,
        begin: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class AnimateTransform(CaseTag):
    """Animates transform attributes on an element

    Args:
        children: Tags, strings, or other rendered content.
        type: Transformation type (rotate|scale|translate|skew).
        by: Relative animation value.
        from_: Starting transformation value.
        to: Ending transformation value.
        dur: Total animation duration.
        repeatCount: Number of repetitions.
        begin: Animation start time.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        type: str | None = None,
        by: str | None = None,
        from_: str | None = None,
        to: str | None = None,
        dur: str | None = None,
        repeatCount: str | float | None = None,
        begin: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Circle(CaseTag):
    """Defines a circle

    Args:
        children: Tags, strings, or other rendered content.
        cx: X-coordinate of center.
        cy: Y-coordinate of center.
        r: Radius.
        pathLength: Total circumference length in user units.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        cx: str | float | None = None,
        cy: str | float | None = None,
        r: str | float | None = None,
        pathLength: float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class ClipPath(CaseTag):
    """Defines a clipping path

    Args:
        children: Tags, strings, or other rendered content.
        clipPathUnits: Coordinate system (userSpaceOnUse|objectBoundingBox).
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        clipPathUnits: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Defs(CaseTag):
    """Defines reusable objects

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Desc(CaseTag):
    """Defines a description of an element

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Ellipse(CaseTag):
    """Defines an ellipse

    Args:
        children: Tags, strings, or other rendered content.
        cx: X-coordinate of center.
        cy: Y-coordinate of center.
        rx: Horizontal radius.
        ry: Vertical radius.
        pathLength: Total path length in user units.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        cx: str | float | None = None,
        cy: str | float | None = None,
        rx: str | float | None = None,
        ry: str | float | None = None,
        pathLength: float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeBlend(CaseTag):
    """Defines image blending

    Args:
        children: Tags, strings, or other rendered content.
        in_: Input image reference.
        in2: Second input image reference.
        mode: Blending mode.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        in_: str | None = None,
        in2: str | None = None,
        mode: str | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeColorMatrix(CaseTag):
    """Applies a matrix transformation on color values

    Args:
        children: Tags, strings, or other rendered content.
        in_: Input image reference.
        type: Matrix type (matrix|saturate|hueRotate|luminanceToAlpha).
        values: Matrix values.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        in_: str | None = None,
        type: str | None = None,
        values: str | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeComponentTransfer(CaseTag):
    """Performs component-wise remapping of data

    Args:
        children: Tags, strings, or other rendered content.
        in_: Input image reference.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        in_: str | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeComposite(CaseTag):
    """Performs image compositing

    Args:
        children: Tags, strings, or other rendered content.
        in_: Input image reference.
        in2: Second input image reference.
        operator: Compositing operation.
        k1: Coefficient for arithmetic operation.
        k2: Coefficient for arithmetic operation.
        k3: Coefficient for arithmetic operation.
        k4: Coefficient for arithmetic operation.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        in_: str | None = None,
        in2: str | None = None,
        operator: str | None = None,
        k1: float | None = None,
        k2: float | None = None,
        k3: float | None = None,
        k4: float | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeConvolveMatrix(CaseTag):
    """Applies a matrix convolution filter

    Args:
        children: Tags, strings, or other rendered content.
        in_: Input image reference.
        order: Matrix dimensions.
        kernelMatrix: Matrix values.
        divisor: Divisor for matrix sum.
        bias: Bias value.
        targetX: Target X position.
        targetY: Target Y position.
        edgeMode: Edge handling mode.
        preserveAlpha: Preserve alpha channel.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        in_: str | None = None,
        order: str | None = None,
        kernelMatrix: str | None = None,
        divisor: float | None = None,
        bias: float | None = None,
        targetX: int | None = None,
        targetY: int | None = None,
        edgeMode: str | None = None,
        preserveAlpha: str | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeDiffuseLighting(CaseTag):
    """Lights an image using diffuse lighting

    Args:
        children: Tags, strings, or other rendered content.
        in_: Input image reference.
        surfaceScale: Surface height scale.
        diffuseConstant: Diffuse lighting constant.
        kernelUnitLength: Kernel unit length.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        in_: str | None = None,
        surfaceScale: float | None = None,
        diffuseConstant: float | None = None,
        kernelUnitLength: str | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeDisplacementMap(CaseTag):
    """Displaces an image using another image

    Args:
        children: Tags, strings, or other rendered content.
        in_: Input image reference.
        in2: Displacement map reference.
        scale: Displacement scale factor.
        xChannelSelector: X displacement channel (R|G|B|A).
        yChannelSelector: Y displacement channel (R|G|B|A).
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        in_: str | None = None,
        in2: str | None = None,
        scale: float | None = None,
        xChannelSelector: str | None = None,
        yChannelSelector: str | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeDistantLight(CaseTag):
    """Defines a distant light source

    Args:
        children: Tags, strings, or other rendered content.
        azimuth: Direction angle on XY plane (degrees).
        elevation: Direction angle from XY plane to z-axis (degrees).
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        azimuth: str | float | None = None,
        elevation: str | float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeDropShadow(CaseTag):
    """Creates a drop shadow effect

    Args:
        children: Tags, strings, or other rendered content.
        dx: X offset of drop shadow.
        dy: Y offset of drop shadow.
        stdDeviation: Blur standard deviation.
        flood_color: Shadow color.
        flood_opacity: Shadow opacity.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        dx: str | float | None = None,
        dy: str | float | None = None,
        stdDeviation: str | float | None = None,
        flood_color: str | None = None,
        flood_opacity: str | float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeFlood(CaseTag):
    """Fills the filter region with a color

    Args:
        children: Tags, strings, or other rendered content.
        flood_color: Fill color.
        flood_opacity: Fill opacity.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        flood_color: str | None = None,
        flood_opacity: str | float | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeFuncA(CaseTag):
    """Defines the alpha transfer function

    Args:
        children: Tags, strings, or other rendered content.
        type: Transfer function type.
        tableValues: Lookup table values.
        slope: Linear function slope.
        intercept: Linear function intercept.
        amplitude: Gamma function amplitude.
        exponent: Gamma function exponent.
        offset: Gamma function offset.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        type: str | None = None,
        tableValues: str | None = None,
        slope: float | None = None,
        intercept: float | None = None,
        amplitude: float | None = None,
        exponent: float | None = None,
        offset: float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeFuncB(CaseTag):
    """Defines the blue transfer function

    Args:
        children: Tags, strings, or other rendered content.
        type: Transfer function type.
        tableValues: Lookup table values.
        slope: Linear function slope.
        intercept: Linear function intercept.
        amplitude: Gamma function amplitude.
        exponent: Gamma function exponent.
        offset: Gamma function offset.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        type: str | None = None,
        tableValues: str | None = None,
        slope: float | None = None,
        intercept: float | None = None,
        amplitude: float | None = None,
        exponent: float | None = None,
        offset: float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeFuncG(CaseTag):
    """Defines the green transfer function

    Args:
        children: Tags, strings, or other rendered content.
        type: Transfer function type.
        tableValues: Lookup table values.
        slope: Linear function slope.
        intercept: Linear function intercept.
        amplitude: Gamma function amplitude.
        exponent: Gamma function exponent.
        offset: Gamma function offset.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        type: str | None = None,
        tableValues: str | None = None,
        slope: float | None = None,
        intercept: float | None = None,
        amplitude: float | None = None,
        exponent: float | None = None,
        offset: float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeFuncR(CaseTag):
    """Defines the red transfer function

    Args:
        children: Tags, strings, or other rendered content.
        type: Transfer function type.
        tableValues: Lookup table values.
        slope: Linear function slope.
        intercept: Linear function intercept.
        amplitude: Gamma function amplitude.
        exponent: Gamma function exponent.
        offset: Gamma function offset.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        type: str | None = None,
        tableValues: str | None = None,
        slope: float | None = None,
        intercept: float | None = None,
        amplitude: float | None = None,
        exponent: float | None = None,
        offset: float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeGaussianBlur(CaseTag):
    """Applies Gaussian blur to an image

    Args:
        children: Tags, strings, or other rendered content.
        in_: Input image reference.
        stdDeviation: Blur amount using bell-curve.
        edgeMode: Edge handling during blur.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        in_: str | None = None,
        stdDeviation: str | float | None = None,
        edgeMode: str | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeImage(CaseTag):
    """Refers to an external image

    Args:
        children: Tags, strings, or other rendered content.
        href: URL to image file.
        preserveAspectRatio: Image scaling control.
        crossorigin: CORS credentials flag.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        href: str | None = None,
        preserveAspectRatio: str | None = None,
        crossorigin: str | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeMerge(CaseTag):
    """Merges multiple filter nodes

    Args:
        children: Tags, strings, or other rendered content.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeMergeNode(CaseTag):
    """Defines a node for feMerge

    Args:
        children: Tags, strings, or other rendered content.
        in_: Input image reference.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        in_: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeMorphology(CaseTag):
    """Applies morphological operations

    Args:
        children: Tags, strings, or other rendered content.
        in_: Input image reference.
        operator: Morphology operator (erode|dilate).
        radius: Morphology radius.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        in_: str | None = None,
        operator: str | None = None,
        radius: str | float | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeOffset(CaseTag):
    """Offsets an image

    Args:
        children: Tags, strings, or other rendered content.
        in_: Input graphic reference.
        dx: Horizontal offset distance.
        dy: Vertical offset distance.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        in_: str | None = None,
        dx: str | float | None = None,
        dy: str | float | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FePointLight(CaseTag):
    """Defines a point light source

    Args:
        children: Tags, strings, or other rendered content.
        x: X-coordinate of light position.
        y: Y-coordinate of light position.
        z: Z-coordinate of light position.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        x: str | float | None = None,
        y: str | float | None = None,
        z: str | float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeSpecularLighting(CaseTag):
    """Lights an image using specular lighting

    Args:
        children: Tags, strings, or other rendered content.
        in_: Input image reference.
        surfaceScale: Surface height scale.
        specularConstant: Specular lighting constant.
        specularExponent: Specular lighting exponent.
        kernelUnitLength: Kernel unit length.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        in_: str | None = None,
        surfaceScale: float | None = None,
        specularConstant: float | None = None,
        specularExponent: float | None = None,
        kernelUnitLength: str | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeSpotLight(CaseTag):
    """Defines a spot light source

    Args:
        children: Tags, strings, or other rendered content.
        x: X-coordinate of light position.
        y: Y-coordinate of light position.
        z: Z-coordinate of light position.
        pointsAtX: X-coordinate of point light points at.
        pointsAtY: Y-coordinate of point light points at.
        pointsAtZ: Z-coordinate of point light points at.
        specularExponent: Focus control for light source.
        limitingConeAngle: Angle of spot light cone.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        x: str | float | None = None,
        y: str | float | None = None,
        z: str | float | None = None,
        pointsAtX: str | float | None = None,
        pointsAtY: str | float | None = None,
        pointsAtZ: str | float | None = None,
        specularExponent: float | None = None,
        limitingConeAngle: float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeTile(CaseTag):
    """Tiles an image to fill a rectangle

    Args:
        children: Tags, strings, or other rendered content.
        in_: Input image reference.
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        in_: str | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class FeTurbulence(CaseTag):
    """Creates turbulence noise

    Args:
        children: Tags, strings, or other rendered content.
        baseFrequency: Base frequency for turbulence.
        numOctaves: Number of noise octaves.
        seed: Random seed for turbulence.
        stitchTiles: Tile stitching mode (stitch|noStitch).
        type: Turbulence type (fractalNoise|turbulence).
        result: Result identifier.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        baseFrequency: str | float | None = None,
        numOctaves: int | None = None,
        seed: float | None = None,
        stitchTiles: str | None = None,
        type: str | None = None,
        result: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Filter(CaseTag):
    """Defines a filter effect

    Args:
        children: Tags, strings, or other rendered content.
        x: X-coordinate of filter region.
        y: Y-coordinate of filter region.
        width: Width of filter region.
        height: Height of filter region.
        filterUnits: Coordinate system for position/size.
        primitiveUnits: Coordinate system for primitives.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        x: str | float | None = None,
        y: str | float | None = None,
        width: str | float | None = None,
        height: str | float | None = None,
        filterUnits: str | None = None,
        primitiveUnits: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class ForeignObject(CaseTag):
    """Allows inclusion of foreign XML

    Args:
        children: Tags, strings, or other rendered content.
        x: X-coordinate.
        y: Y-coordinate.
        width: Width.
        height: Height.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        x: str | float | None = None,
        y: str | float | None = None,
        width: str | float | None = None,
        height: str | float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class G(CaseTag):
    """Groups SVG elements

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Image(CaseTag):
    """Embeds an image

    Args:
        children: Tags, strings, or other rendered content.
        x: Horizontal position from origin.
        y: Vertical position from origin.
        width: Width (required).
        height: Height (required).
        href: URL to image file.
        preserveAspectRatio: Image scaling control.
        crossorigin: CORS credentials flag.
        decoding: Image decoding hint.
        fetchpriority: Fetch priority hint (experimental).
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        x: str | float | None = None,
        y: str | float | None = None,
        width: str | float | None = None,
        height: str | float | None = None,
        href: str | None = None,
        preserveAspectRatio: str | None = None,
        crossorigin: str | None = None,
        decoding: str | None = None,
        fetchpriority: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Line(CaseTag):
    """Defines a line

    Args:
        children: Tags, strings, or other rendered content.
        x1: X-coordinate of start point.
        y1: Y-coordinate of start point.
        x2: X-coordinate of end point.
        y2: Y-coordinate of end point.
        pathLength: Total path length in user units.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        x1: str | float | None = None,
        y1: str | float | None = None,
        x2: str | float | None = None,
        y2: str | float | None = None,
        pathLength: float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class LinearGradient(CaseTag):
    """Defines a linear gradient

    Args:
        children: Tags, strings, or other rendered content.
        x1: X-coordinate of gradient start.
        y1: Y-coordinate of gradient start.
        x2: X-coordinate of gradient end.
        y2: Y-coordinate of gradient end.
        gradientUnits: Coordinate system.
        gradientTransform: Additional transformation.
        href: Reference to template gradient.
        spreadMethod: Gradient behavior outside bounds.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        x1: str | float | None = None,
        y1: str | float | None = None,
        x2: str | float | None = None,
        y2: str | float | None = None,
        gradientUnits: str | None = None,
        gradientTransform: str | None = None,
        href: str | None = None,
        spreadMethod: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Marker(CaseTag):
    """Defines a graphic for drawing on lines

    Args:
        children: Tags, strings, or other rendered content.
        markerWidth: Width of marker viewport.
        markerHeight: Height of marker viewport.
        markerUnits: Coordinate system.
        refX: X reference point.
        refY: Y reference point.
        orient: Marker orientation.
        viewBox: Viewport bounds.
        preserveAspectRatio: Aspect ratio handling.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        markerWidth: str | float | None = None,
        markerHeight: str | float | None = None,
        markerUnits: str | None = None,
        refX: str | float | None = None,
        refY: str | float | None = None,
        orient: str | float | None = None,
        viewBox: str | None = None,
        preserveAspectRatio: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Mask(CaseTag):
    """Defines a mask

    Args:
        children: Tags, strings, or other rendered content.
        x: X-coordinate of top-left corner.
        y: Y-coordinate of top-left corner.
        width: Width of masking area.
        height: Height of masking area.
        maskUnits: Coordinate system for position/size.
        maskContentUnits: Coordinate system for contents.
        mask_type: Mask mode (alpha|luminance).
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        x: str | float | None = None,
        y: str | float | None = None,
        width: str | float | None = None,
        height: str | float | None = None,
        maskUnits: str | None = None,
        maskContentUnits: str | None = None,
        mask_type: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Metadata(CaseTag):
    """Defines metadata

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Mpath(CaseTag):
    """Defines a motion path

    Args:
        children: Tags, strings, or other rendered content.
        href: Reference to path element.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        href: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Path(CaseTag):
    """Defines a path

    Args:
        children: Tags, strings, or other rendered content.
        d: Path data defining the shape.
        pathLength: Total path length in user units.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        d: str | None = None,
        pathLength: float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Pattern(CaseTag):
    """Defines a pattern

    Args:
        children: Tags, strings, or other rendered content.
        x: X-coordinate shift of pattern tile.
        y: Y-coordinate shift of pattern tile.
        width: Width of pattern tile.
        height: Height of pattern tile.
        patternUnits: Coordinate system for position/size.
        patternContentUnits: Coordinate system for contents.
        patternTransform: Additional transformation.
        href: Reference to template pattern.
        viewBox: Viewport bounds for pattern.
        preserveAspectRatio: Aspect ratio handling.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        x: str | float | None = None,
        y: str | float | None = None,
        width: str | float | None = None,
        height: str | float | None = None,
        patternUnits: str | None = None,
        patternContentUnits: str | None = None,
        patternTransform: str | None = None,
        href: str | None = None,
        viewBox: str | None = None,
        preserveAspectRatio: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Polygon(CaseTag):
    """Defines a polygon

    Args:
        children: Tags, strings, or other rendered content.
        points: List of x,y coordinate pairs.
        pathLength: Total path length in user units.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        points: str | None = None,
        pathLength: float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Polyline(CaseTag):
    """Defines a polyline

    Args:
        children: Tags, strings, or other rendered content.
        points: List of x,y coordinate pairs.
        pathLength: Total path length in user units.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        points: str | None = None,
        pathLength: float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class RadialGradient(CaseTag):
    """Defines a radial gradient

    Args:
        children: Tags, strings, or other rendered content.
        cx: X-coordinate of end circle.
        cy: Y-coordinate of end circle.
        r: Radius of end circle.
        fx: X-coordinate of start circle.
        fy: Y-coordinate of start circle.
        fr: Radius of start circle.
        gradientUnits: Coordinate system.
        gradientTransform: Additional transformation.
        href: Reference to template gradient.
        spreadMethod: Gradient behavior.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        cx: str | float | None = None,
        cy: str | float | None = None,
        r: str | float | None = None,
        fx: str | float | None = None,
        fy: str | float | None = None,
        fr: str | float | None = None,
        gradientUnits: str | None = None,
        gradientTransform: str | None = None,
        href: str | None = None,
        spreadMethod: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Rect(CaseTag):
    """Defines a rectangle

    Args:
        children: Tags, strings, or other rendered content.
        x: X-coordinate.
        y: Y-coordinate.
        width: Width.
        height: Height.
        rx: Horizontal corner radius.
        ry: Vertical corner radius.
        pathLength: Total perimeter length in user units.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        x: str | float | None = None,
        y: str | float | None = None,
        width: str | float | None = None,
        height: str | float | None = None,
        rx: str | float | None = None,
        ry: str | float | None = None,
        pathLength: float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Script(CaseTag):
    """Defines a script

    Args:
        children: Tags, strings, or other rendered content.
        type: Script MIME type.
        href: External script URL.
        crossorigin: CORS credentials flag.
        fetchpriority: Fetch priority hint (experimental).
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        type: str | None = None,
        href: str | None = None,
        crossorigin: str | None = None,
        fetchpriority: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Set(CaseTag):
    """Sets an attribute value

    Args:
        children: Tags, strings, or other rendered content.
        to: Value to apply for animation duration.
        attributeName: Target attribute to set.
        begin: Animation start time.
        dur: Animation duration.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        to: str | None = None,
        attributeName: str | None = None,
        begin: str | None = None,
        dur: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Stop(CaseTag):
    """Defines a gradient stop

    Args:
        children: Tags, strings, or other rendered content.
        offset: Position along gradient vector.
        stop_color: Color of gradient stop.
        stop_opacity: Opacity of gradient stop.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        offset: str | float | None = None,
        stop_color: str | None = None,
        stop_opacity: str | float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Style(CaseTag):
    """Defines style information

    Args:
        children: Tags, strings, or other rendered content.
        type: Style sheet language MIME type.
        media: Media query for when styles apply.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        type: str | None = None,
        media: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Svg(CaseTag):
    """Defines an SVG document fragment

    Args:
        children: Tags, strings, or other rendered content.
        width: Displayed width of viewport.
        height: Displayed height of viewport.
        x: X-coordinate of container.
        y: Y-coordinate of container.
        viewBox: SVG viewport coordinates.
        preserveAspectRatio: Aspect ratio handling.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        width: str | float | None = None,
        height: str | float | None = None,
        x: str | float | None = None,
        y: str | float | None = None,
        viewBox: str | None = None,
        preserveAspectRatio: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Switch(CaseTag):
    """Defines conditional processing

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Symbol(CaseTag):
    """Defines a reusable symbol

    Args:
        children: Tags, strings, or other rendered content.
        width: Width of symbol.
        height: Height of symbol.
        x: X-coordinate.
        y: Y-coordinate.
        viewBox: Viewport bounds for symbol.
        preserveAspectRatio: Aspect ratio handling.
        refX: X reference point.
        refY: Y reference point.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        width: str | float | None = None,
        height: str | float | None = None,
        x: str | float | None = None,
        y: str | float | None = None,
        viewBox: str | None = None,
        preserveAspectRatio: str | None = None,
        refX: str | float | None = None,
        refY: str | float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Text(CaseTag):
    """Defines text content

    Args:
        children: Tags, strings, or other rendered content.
        x: X-coordinates of text baseline.
        y: Y-coordinates of text baseline.
        dx: Horizontal shift from previous text.
        dy: Vertical shift from previous text.
        rotate: Rotation of individual glyphs.
        lengthAdjust: Text stretching method.
        textLength: Target width for text scaling.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        x: str | float | None = None,
        y: str | float | None = None,
        dx: str | float | None = None,
        dy: str | float | None = None,
        rotate: str | None = None,
        lengthAdjust: str | None = None,
        textLength: str | float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class TextPath(CaseTag):
    """Defines text along a path

    Args:
        children: Tags, strings, or other rendered content.
        href: Reference to path element for text layout.
        lengthAdjust: Length adjustment method.
        method: Glyph rendering method.
        path: Path data for text layout.
        side: Which side of path to render text.
        spacing: Glyph spacing handling.
        startOffset: Offset from path beginning.
        textLength: Text rendering width.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        href: str | None = None,
        lengthAdjust: str | None = None,
        method: str | None = None,
        path: str | None = None,
        side: str | None = None,
        spacing: str | None = None,
        startOffset: str | float | None = None,
        textLength: str | float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Title(CaseTag):
    """Defines a title for the SVG document

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Tspan(CaseTag):
    """Defines a text span

    Args:
        children: Tags, strings, or other rendered content.
        x: X-coordinates of text baseline.
        y: Y-coordinates of text baseline.
        dx: Horizontal shift from previous text.
        dy: Vertical shift from previous text.
        rotate: Rotation of individual glyphs.
        lengthAdjust: Text stretching method.
        textLength: Target width for text scaling.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        x: str | float | None = None,
        y: str | float | None = None,
        dx: str | float | None = None,
        dy: str | float | None = None,
        rotate: str | None = None,
        lengthAdjust: str | None = None,
        textLength: str | float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Use(CaseTag):
    """References another element

    Args:
        children: Tags, strings, or other rendered content.
        href: Reference to element to duplicate.
        x: X offset transformation.
        y: Y offset transformation.
        width: Width (only for elements with viewBox).
        height: Height (only for elements with viewBox).
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        href: str | None = None,
        x: str | float | None = None,
        y: str | float | None = None,
        width: str | float | None = None,
        height: str | float | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class View(CaseTag):
    """Defines a view

    Args:
        children: Tags, strings, or other rendered content.
        viewBox: Viewport bounds.
        preserveAspectRatio: Aspect ratio handling.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        **kwargs: Additional attributes.
    """

    def __init__(
        self,
        *children: Any,
        viewBox: str | None = None,
        preserveAspectRatio: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))



"""Root module for the Air Tags system."""

from __future__ import annotations

import html
import json
from collections.abc import Mapping
from functools import cached_property
from types import MappingProxyType
from typing import Any, ClassVar, Final, Self, TypedDict

from ..utils import SafeStr, clean_html_attr_key, pretty_format_html, pretty_print_html

type AttributesType = str | int | float | bool


class TagDictType(TypedDict):
    name: str
    attributes: dict[str, AttributesType]
    children: tuple[Any, ...]


class TagKeys:
    NAME: Final = "name"
    ATTRIBUTES: Final = "attributes"
    CHILDREN: Final = "children"


class BaseTag:
    """Base tag for all other tags.

    Sets four attributes, name, module, children, and attrs.
    These are important for Starlette view responses, as nested objects
    get auto-serialized to JSON and need to be rebuilt. With
    the values of these attributes, the object reconstruction can occur.
    """

    _registry: ClassVar[dict[str, type[BaseTag]]] = {}
    registry: ClassVar[Mapping[str, type[BaseTag]]] = MappingProxyType(_registry)  # read-only view

    def __init__(self, *children: Any, **kwargs: AttributesType) -> None:
        """
        Args:
            children: Tags, strings, or other rendered content.
            kwargs: Keyword arguments transformed into tag attributes.
        """
        self._name = self.__class__.__name__
        self._module = self.__class__.__module__
        self._children, self._attrs = children, kwargs

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        """Non-instantiable base; all subclasses are instantiable."""
        if cls is BaseTag:
            msg = f"{cls.__name__} cannot be instantiated; use a subclass"
            raise TypeError(msg)
        return super().__new__(cls)

    @property
    def name(self) -> str:
        return self._name.lower()

    @cached_property
    def attrs(self) -> str:
        if not self._attrs:
            return ""
        return " " + " ".join(self._format_attr(key) for key, value in self._attrs.items() if value is not False)

    def _format_attr(self, key: str) -> str:
        value = self._attrs[key]
        clean_key = clean_html_attr_key(key)
        if value is True:
            return clean_key
        return f'{clean_key}="{value}"'

    @cached_property
    def children(self) -> str:
        if not self._children:
            return ""
        return "".join(self._render_child(child) for child in self._children)

    def _render_child(self, child: Any) -> str:
        child_str = str(child)
        if isinstance(child, BaseTag | SafeStr):
            return child_str
        return self._escape_text(child_str)

    @staticmethod
    def _escape_text(text: str) -> str:
        return html.escape(text)

    def render(self) -> str:
        return self._render()

    def pretty_render(
        self,
        *,
        with_body: bool = False,
        with_head: bool = False,
        with_doctype: bool = False,
    ) -> str:
        """Pretty-print without escaping."""
        return pretty_format_html(self._render(), with_body=with_body, with_head=with_head, with_doctype=with_doctype)

    def pretty_print(self) -> None:
        """Pretty-print and render HTML with syntax highlighting."""
        pretty_print_html(self.pretty_render())

    def _render(self) -> str:
        return self._render_paired()

    def _render_void(self) -> str:
        """Render a self-closing (void) tag.

        TODO: HTML5 does not require a trailing slash.
              Consider returning f"<{self.name}{self.attrs}>" instead.
        """
        return f"<{self.name}{self.attrs} />"

    def _render_paired(self) -> str:
        """Render a normal open/close pair."""
        return f"<{self.name}{self.attrs}>{self.children}</{self.name}>"

    def __str__(self) -> str:
        return self.render()

    def __repr__(self) -> str:
        summary = f'("{self._doc_summary}")' if self._doc_summary else ""
        return f"<air.{self._name}{summary}>"

    @property
    def _doc_summary(self) -> str:
        return self.__doc__.splitlines()[0] if self.__doc__ else ""

    def full_repr(self) -> str:
        attributes = f"{TagKeys.ATTRIBUTES}={self._attrs}" if self._attrs else ""
        children = ", ".join(child.full_repr() if isinstance(child, BaseTag) else child for child in self._children)
        children_str = f"{attributes and ', '}{TagKeys.CHILDREN}={children}" if self._children else ""
        return f"{self._name}({attributes}{children_str})"

    def to_pretty_dict(self) -> str:
        try:
            from rich.pretty import pretty_repr

            return pretty_repr(self.to_dict(), max_width=170, max_length=7, max_depth=4, max_string=25)
        except ModuleNotFoundError:
            return str(self.to_dict())

    def to_dict(self) -> TagDictType:
        """
        Convert this Tag into plain Python data that json.dumps can serialize.

        - Children that are Tag instances are converted recursively.
        """
        return {
            TagKeys.NAME: self._name,
            TagKeys.ATTRIBUTES: self._attrs,
            TagKeys.CHILDREN: tuple(self.to_child_dict()),
        }

    def to_child_dict(self) -> list[TagDictType]:
        return [child.to_dict() if isinstance(child, BaseTag) else child for child in self._children]

    def to_json(self, indent: int | None = None) -> str:
        return json.dumps(self.to_dict(), ensure_ascii=False, indent=indent)

    def to_pretty_json(self) -> str:
        return self.to_json(indent=4)

    @classmethod
    def from_dict(cls, source_dict: TagDictType) -> Self:
        name, attributes, children_dict = source_dict.values()
        children: tuple[Self, ...] = tuple(cls._from_child_dict(children_dict))
        tag = cls.registry[name](*children, **attributes)
        tag._name = name
        return tag

    @classmethod
    def _from_child_dict(cls, children_dict: TagDictType) -> list[Self]:
        return [
            cls.from_dict(child_dict) if isinstance(child_dict, dict) else child_dict for child_dict in children_dict
        ]

    @classmethod
    def from_json(cls, source_json: str) -> Self:
        return cls.from_dict(json.loads(source_json))

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        BaseTag._registry[cls.__name__] = cls



from __future__ import annotations

from typing import Final

from .models import (
    Base,
    BaseTag,
    Link,
    Meta,
    Script,
    Style,
    Title,
)
from .models.base import AttributesType as AttributesType

type TagTypes = tuple[type[BaseTag], ...]
HEAD_TAG_TYPES: Final[TagTypes] = (Title, Style, Meta, Link, Script, Base)



"""A FastAPI-powered breath of fresh air in Python web development."""

from starlette.staticfiles import StaticFiles as StaticFiles

from . import (
    ext as ext,
    layouts as layouts,
    responses as responses,
)
from .applications import Air as Air
from .background import BackgroundTasks as BackgroundTasks
from .dependencies import is_htmx_request as is_htmx_request
from .exceptions import (
    HTTPException as HTTPException,
    RenderException as RenderException,
)
from .forms import (
    AirField as AirField,
    AirForm as AirForm,
)
from .middleware import SessionMiddleware as SessionMiddleware
from .requests import Request as Request
from .responses import (
    AirResponse as AirResponse,
    RedirectResponse as RedirectResponse,
    SSEResponse as SSEResponse,
    TagResponse as TagResponse,
)
from .routing import AirRouter as AirRouter
from .tags import (
    H1 as H1,
    H2 as H2,
    H3 as H3,
    H4 as H4,
    H5 as H5,
    H6 as H6,
    A as A,
    Abbr as Abbr,
    Address as Address,
    Area as Area,
    Article as Article,
    Aside as Aside,
    Audio as Audio,
    B as B,
    Base as Base,
    BaseTag as BaseTag,
    Bdi as Bdi,
    Bdo as Bdo,
    Blockquote as Blockquote,
    Body as Body,
    Br as Br,
    Button as Button,
    Canvas as Canvas,
    Caption as Caption,
    Children as Children,
    Cite as Cite,
    Code as Code,
    Col as Col,
    Colgroup as Colgroup,
    Data as Data,
    Datalist as Datalist,
    Dd as Dd,
    Del as Del,
    Details as Details,
    Dfn as Dfn,
    Dialog as Dialog,
    Div as Div,
    Dl as Dl,
    Dt as Dt,
    Em as Em,
    Embed as Embed,
    Fieldset as Fieldset,
    Figcaption as Figcaption,
    Figure as Figure,
    Footer as Footer,
    Form as Form,
    Fragment as Fragment,
    Head as Head,
    Header as Header,
    Hgroup as Hgroup,
    Hr as Hr,
    Html as Html,
    I as I,
    Iframe as Iframe,
    Img as Img,
    Input as Input,
    Ins as Ins,
    Kbd as Kbd,
    Label as Label,
    Legend as Legend,
    Li as Li,
    Link as Link,
    Main as Main,
    Map as Map,
    Mark as Mark,
    Menu as Menu,
    Meta as Meta,
    Meter as Meter,
    Nav as Nav,
    Noscript as Noscript,
    Object as Object,
    Ol as Ol,
    Optgroup as Optgroup,
    Option as Option,
    Output as Output,
    P as P,
    Param as Param,
    Picture as Picture,
    Pre as Pre,
    Progress as Progress,
    Q as Q,
    Raw as Raw,
    Rp as Rp,
    Rt as Rt,
    Ruby as Ruby,
    S as S,
    SafeStr as SafeStr,
    Samp as Samp,
    Script as Script,
    Search as Search,
    Section as Section,
    Select as Select,
    SelfClosingTag as SelfClosingTag,
    Small as Small,
    Source as Source,
    Span as Span,
    Strong as Strong,
    Style as Style,
    Sub as Sub,
    Summary as Summary,
    Sup as Sup,
    Table as Table,
    Tag as Tag,
    TagDictType as TagDictType,
    TagKeys as TagKeys,
    Tags as Tags,
    Tbody as Tbody,
    Td as Td,
    Template as Template,
    Textarea as Textarea,
    Tfoot as Tfoot,
    Th as Th,
    Thead as Thead,
    Time as Time,
    Title as Title,
    Tr as Tr,
    Track as Track,
    Transparent as Transparent,
    U as U,
    Ul as Ul,
    UnSafeTag as UnSafeTag,
    Var as Var,
    Video as Video,
    Wbr as Wbr,
    svg as svg,
)
from .templating import (
    JinjaRenderer as JinjaRenderer,
    Renderer as Renderer,
)



"""Background tasks in Air, for those times a long-running process is needed that doesn't force the user to wait."""

from collections.abc import Callable
from typing import Annotated

from fastapi.background import BackgroundTasks as FastAPIBackgroundTasks
from typing_extensions import Doc


class BackgroundTasks(FastAPIBackgroundTasks):
    """
    A collection of background tasks that will be called after a response has been
    sent to the client.

    Example:

        from air import BackgroundTasks, Air

        app = Air()


        def write_notification(email: str, message=""):
            with open("log.txt", mode="w") as email_file:
                content = f"notification for {email}: {message}"
                email_file.write(content)


        @app.post("/send-notification/{email}")
        async def send_notification(email: str, background_tasks: BackgroundTasks):
            message = "some notification"
            background_tasks.add_task(write_notification, email, message=message)
            content = f"notification for {email}: {message}"
            return air.layouts.picocss(air.Title(content), air.H1(content))

    """

    def add_task[**P, T](
        self,
        func: Annotated[
            Callable[P, T],
            Doc(
                """
                The function to call after the response is sent.

                It can be a regular `def` function or an `async def` function.
                """
            ),
        ],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """
        Add a function to be called in the background after the response is sent.
        """
        return super().add_task(func, *args, **kwargs)



from __future__ import annotations

from typing import Final

from fastapi.datastructures import Default
from fastapi.utils import generate_unique_id

from .responses import AirResponse, JSONResponse

# Public default so apps/routers can import the same sentinel (meaning: special marker)
default_html_response: Final = Default(AirResponse)
default_json_response: Final = Default(JSONResponse)
default_generate_unique_id: Final = Default(generate_unique_id)


def compute_page_path(endpoint_name: str) -> str:
    """index -> '/', otherwise '/name-with-dashes'."""
    return "/" if endpoint_name == "index" else f"/{endpoint_name.replace('_', '-')}"



"""Special Air Tags that aren't find in any other category."""

from typing import Any, Literal, override

from ..utils import HTML_DOCTYPE, locals_cleanup
from .base import AttributesType, BaseTag


class Html(BaseTag):
    """Defines the root of an HTML document"""

    @override
    def _render(self) -> str:
        return f"{HTML_DOCTYPE}{self._render_paired()}"

    @override
    def pretty_render(
        self,
        *,
        with_body: bool = False,
        with_head: bool = False,
        with_doctype: bool = True,
    ) -> str:
        """Pretty-print without escaping."""
        return super().pretty_render(with_body=with_body, with_head=with_head, with_doctype=with_doctype)


class Transparent(BaseTag):
    """Transparent(no own HTML tag) container that renders only its children(just the inner content)."""

    def __init__(
        self,
        *children: Any,
    ) -> None:
        super().__init__(*children)

    @override
    def _render(self) -> str:
        return self.children


class Children(Transparent):
    pass


class Tag(Transparent):
    pass


class Tags(Transparent):
    pass


class Fragment(Transparent):
    pass


class SelfClosingTag(BaseTag):
    def __init__(
        self,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs)

    @override
    def _render(self) -> str:
        return self._render_void()


class UnSafeTag(BaseTag):
    @override
    def __init__(self, text_child: str = "", /, **kwargs: AttributesType) -> None:
        super().__init__(text_child, **kwargs)
        if not isinstance(text_child, str):
            msg = f"{self!r} only accepts string content"
            raise TypeError(msg)

    @override
    @staticmethod
    def _escape_text(text: str) -> str:
        return text


class Raw(UnSafeTag, Transparent):
    """Renders raw HTML content without escaping.

    Raises:
        TypeError: If non-string content is provided
    Example:
        Raw('<strong>Bold</strong> text')
        # Produces '<strong>Bold</strong> text'
        # Use with other tags
        Div(
            P("Safe content"),
            Raw('<hr class="divider">'),
            P("More safe content")
        )
    """


class Script(UnSafeTag):
    """Defines a client-side script.
    Warning: Script tag does not protect against code injection.

    Args:
        text_child: Inline script code. Use an empty string when providing ``src``.
        src: URI of the external script.
        type: Script type. Examples: ``module``, ``importmap``, ``speculationrules``,
            a JavaScript MIME type (e.g. ``text/javascript``), or empty for classic scripts.
        async_: Fetch in parallel and execute as soon as ready; order is not guaranteed.
        defer: Execute after parsing (classic scripts only; modules defer by default).
        nomodule: Do not execute on browsers that support ES modules.
        crossorigin: CORS mode. One of ``"anonymous"`` or ``"use-credentials"``.
        integrity: Subresource Integrity hash (e.g. ``"sha384-..."``).
        referrerpolicy: Which referrer to send. One of:
            ``"no-referrer"``, ``"no-referrer-when-downgrade"``, ``"origin"``,
            ``"origin-when-cross-origin"``, ``"same-origin"``, ``"strict-origin"``,
            ``"strict-origin-when-cross-origin"``, ``"unsafe-url"``.
        fetchpriority: Network priority hint. One of ``"high"``, ``"low"``, ``"auto"``.
        blocking: Space-separated tokens that block operations; currently ``"render"``.
        attributionsrc: Space-separated URLs for Attribution Reporting (experimental).
        nonce: CSP nonce (meaning: one-time token) to allow this inline script.
        class_: Substituted as the DOM ``class`` attribute.
        id: DOM ``id`` attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    @override
    def __init__(
        self,
        text_child: str = "",
        /,
        *,
        src: str | None = None,
        type: str | None = None,
        async_: bool = False,
        defer: bool = False,
        nomodule: bool = False,
        crossorigin: Literal["anonymous", "use-credentials"] | None = None,
        integrity: str | None = None,
        referrerpolicy: Literal[
            "no-referrer",
            "no-referrer-when-downgrade",
            "origin",
            "origin-when-cross-origin",
            "same-origin",
            "strict-origin",
            "strict-origin-when-cross-origin",
            "unsafe-url",
        ]
        | None = None,
        fetchpriority: Literal["high", "low", "auto"] | None = None,
        blocking: Literal["render"] | None = None,
        attributionsrc: str | None = None,
        nonce: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(text_child, **kwargs | locals_cleanup(locals()))


class Style(UnSafeTag):
    """Defines style information for a document.
    Warning: Style tag does not protect against code injection.

    Args:
        text_child: CSS stylesheet text.
        media: Media query (e.g. ``"(width < 500px)"``). Defaults to ``"all"``.
        title: Title for alternate style sheet sets.
        blocking: Space-separated tokens that block operations; currently ``"render"``.
        nonce: CSP nonce (meaning: one-time token) to allow this inline style.
        type: (Deprecated) Only ``""`` or ``"text/css"`` are permitted; omit in modern HTML.
        class_: Substituted as the DOM ``class`` attribute.
        id: DOM ``id`` attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    @override
    def __init__(
        self,
        text_child: str = "",
        /,
        *,
        media: str | None = None,
        title: str | None = None,
        blocking: Literal["render"] | None = None,
        nonce: str | None = None,
        type: str | None = None,  # deprecated
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(text_child, **kwargs | locals_cleanup(locals()))
